#include "Field.h"
#include "Action.h"
#include "Site.h"
#include "Space.h"
#include "MASS.h"
using namespace MMASS;

Action::~Action() { }

spAgentInfo_set& Action::getReactionPartners() {
	return reactionPartners;
}

spAgentInfo_set::const_iterator Action::beginReactionPartners() {
	return reactionPartners.begin();
}

spAgentInfo_set::const_iterator Action::endReactionPartners() {
	return reactionPartners.end();
}

const string& Action::getDestinationSiteName() {
	return siteName;
}

spAction Action::getSelf() {
	return self.lock();
}

Action::Action() { }

Action::Action(spField field_) { field = field_; }
Action::Action(const string &siteName_) : siteName(siteName_) { }
Action::Action(spAgentInfo_set &names) : reactionPartners(names) { }

void Action::setSelf(MMASS::wpAction self_) {
	self = self_;
}

Reaction::Reaction(spAgentInfo_set &names) : Action(names) { }

bool Reaction::manage(spAgent agent, spAction action, unsigned long turn) {
	return _reactionManagement(agent, action, turn);
}

spActionInfo Reaction::createActionInfo() {
	return ActionInfo::createActionInfo(this->getIdentifier(), this->reactionPartners);
}

const string& ActionInfo::getIdentifier() {
	return actionIdentifier;
}

spAgentInfo_set::const_iterator ActionInfo::beginReactionPartners() {
	return reactionPartners.begin();
}

spAgentInfo_set::const_iterator ActionInfo::endReactionPartners() {
	return reactionPartners.end();
}

ActionInfo::ActionInfo(const string &actionIdentifier_) : actionIdentifier(actionIdentifier_) { }
ActionInfo::ActionInfo(const string &actionIdentifier_, spField field_) : actionIdentifier(actionIdentifier_) { field = field_; }
ActionInfo::ActionInfo(const string &actionIdentifier_, const string &siteName_) : actionIdentifier(actionIdentifier_), siteName(siteName_) { }
ActionInfo::ActionInfo(const string &actionIdentifier_, spAgentInfo_set &reactionPartners_) : actionIdentifier(actionIdentifier_), reactionPartners(reactionPartners_) { }

spActionInfo ActionInfo::createActionInfo(const string &actionIdentifier_) {
	return spActionInfo(new ActionInfo(actionIdentifier_));
}

spActionInfo ActionInfo::createActionInfo(const string &actionIdentifier_, spField field_) {
	return spActionInfo(new ActionInfo(actionIdentifier_, field_));
}

spActionInfo ActionInfo::createActionInfo(const string &actionIdentifier_, const string &siteName_) {
	return spActionInfo(new ActionInfo(actionIdentifier_, siteName_));
}

spActionInfo ActionInfo::createActionInfo(const string &actionIdentifier_, spAgentInfo_set &reactionPartners_) {
	return spActionInfo(new ActionInfo(actionIdentifier_, reactionPartners_));
}

Transport::~Transport() { }

spAction Transport::createTransport(const string & siteName_) {
	return spAction(new Transport(siteName_));
}

void Transport::performAction(spAgent agent) {
	spSite site = agent->getMASS()->getSpace()->getSite(this->siteName);
	spAgent agentOnSite = site->getAgent();
	if (agentOnSite == agent) return;
	site->setAgentToNull();
	site->setAgent(agent);
}

Transport::Transport(const string &siteName_) : Action(siteName_) { }

bool Transport::manage(spAgent agent, spAction action, unsigned long turn) {
	return _transportManagement(agent, action, turn);
}

spActionInfo Transport::createActionInfo() {
	return ActionInfo::createActionInfo(this->getIdentifier(), this->siteName);
}

Emit::~Emit() { }

spAction Emit::createEmit(spField field_) {
	return spAction(new Emit(field_));
}

void Emit::performAction(spAgent agent) {
	spSpace space = agent->getMASS()->getSpace();
	spSite source = agent->getSite();
	spField field = this->field;
	source->addField(field->createField(field->diffusion(source, field->getValue(), source)));

	queue<const string> sitesInOrder = source->getSitesInDistanceOrder();
	while(!sitesInOrder.empty()) {
		const string siteName = sitesInOrder.front();
		sitesInOrder.pop();
		spSite site = space->getSite(siteName);
		site->addField(field->createField(field->diffusion(source, field->getValue(), site)));
		//printf("%s has field %s = %g\n", site->getName().c_str(), field->getIdentifier().c_str(), site->fields[field->getIdentifier()]->getValue());
	}
}

Emit::Emit(spField field_) : Action(field_) { }

bool Emit::manage(spAgent agent, spAction action, unsigned long turn) {
	return _emitManagement(agent, action, turn);
}

spActionInfo Emit::createActionInfo() {
	return ActionInfo::createActionInfo(this->getIdentifier(), this->field);
}