#ifndef MASS_H
#define MASS_H
#include <OgreNoMemoryMacros.h>
#include "MMASSDefs.h"
#include "Agent.h"
#include <string>
#include <set>
#include <stdexcept>

namespace MMASS {
	using std::string;
	using std::set;

	class MASS {
	public:
		class logic_error : public std::logic_error {
		public:
			logic_error(const string& msg = "") : std::logic_error(msg) {}
		};

		class invalid_argument : public std::invalid_argument {
		public:
			invalid_argument(const string& msg = "") : std::invalid_argument(msg) {}
		};

	protected:
		spSpace     space;  // Space
		spAgent_map agents; // Agents
		// fields

	public:
		~MASS();

		// TODO: CRIAR UM SPACE AUTOMATICAMENTE!!!!!!!!!!!
		spSpace createSpace() throw (logic_error);

		spAgent createAgent(Agent::AgentFactory &factory, const string &name_)
			throw (invalid_argument, logic_error);
		spAgent createAgent(Agent::AgentFactory &factory, const string &agent_name,
			const string site_name) throw (invalid_argument, logic_error);

		const string& getName() const;

		const spMASS getSelf() const;

		const spMMASS getMMASS() const;

		const bool hasAgent() const;
		const bool hasAgent(const string &name_) const;
		const bool hasAgent(const spAgent agent) const;
		spAgent_map::const_iterator beginAgents() const;
		spAgent_map::const_iterator endAgents() const;
		const spAgent getAgentWithoutSite() const;
		const size_t numOfAgents() const;
		spAgent getAgent(const string &name_) throw (logic_error, invalid_argument);

		const bool hasSpace() const;
		const spSpace getSpace() const;

		void finish() throw (logic_error);
		const bool isReady() const;

	private:
		const string      name;
		wpMASS            self;
		wpMMASS           mmass;
		bool              ready;
		set<const string> occupiedSites;

		static spMASS createMASS(const string &name_);

		MASS(const string &name_);

		void setSelfPointer(wpMASS self_);

		void setMMASS(wpMMASS mmass_);

		set<const string>& getOccupiedSites();
		void addOccupiedSite(const string &name_);
		void eraseOccupiedSite(const string &name_);
		void clearOccupiedSites();

		/* Hiding methods. */
		MASS(const MASS&);
		MASS& operator=(const MASS&);

		friend CMMASS;
		friend void _clearOccupiedSites(spMMASS mmass);
		friend bool _transportManagement(spAgent agent, spAction action, unsigned long turn);
		friend bool _reactionManagement(spAgent agent, spAction action, unsigned long turn);
	};
}

#endif