#include "MMASS.h"
#include "Action.h"
#include "Site.h"
#include "Space.h"
#include "ThreadArgument.h"

#include "Field.h" // TEMPORARY

#include <map>
using namespace MMASS;

CMMASS::~CMMASS() {
	int rc;
	rc = pthread_mutex_destroy(&const_cast<pthread_mutex_t>(mutex));
	rc += pthread_cond_destroy(&const_cast<pthread_cond_t>(cond));
	if (rc) printf("CMMASS::~CMMASS() ln8\n");
}

const spMMASS CMMASS::getSingleton() {
	return self;
}

const spMMASS CMMASS::getSelf() const {
	return self;
}

spMASS CMMASS::createMASS(const string &name_) throw (logic_error, invalid_argument) {
	if (ready)
		throw logic_error("CMMASS::createMASS(const string &name_)"
			" - name_: "+name_+"\n"
			"The MMASS is ready. Can't create MASS with a finished MMASS.\n");

	if (this->hasMASS(name_)) {
		throw invalid_argument("CMMASS::createMASS(const string &name_)"
			" - name_: "+name_+"\n"
			"A MASS with this name already exists.\n");
	}

	spMASS mass = MASS::createMASS(name_);
	
	mass->setMMASS(self);
	masss[name_] = mass;
	return mass;
}

spMASS CMMASS::getMASS(const string &name_) throw (logic_error, invalid_argument) {
	if (!this->hasMASS()) {
		throw logic_error(string("CMMASS::getMASS(const string &name_)"
			" - name_: "+name_+"\n"
			"There is no MASS. Create one first before "
			"getting it.\n"));
	}

	if (!this->hasMASS(name_)) {
		throw invalid_argument(string("CMMASS::getMASS(const string &name_)"
			" - name_: "+name_+"\n"
			"There is no MASS with this name.\n"));
	}

	return masss[name_];
}

const bool CMMASS::hasMASS() const {
	return (!masss.empty());
}

const bool CMMASS::hasMASS(const string &name_) const {
	return (masss.count(name_) > 0);
}

spMASS_map::const_iterator CMMASS::beginMASSs() const {
	return masss.begin();
}

spMASS_map::const_iterator CMMASS::endMASSs() const {
	return masss.end();
}

const spMASS CMMASS::getUnfinishedMASS() const {
	spMASS_map::const_iterator mass_i, end = masss.end();
	for (mass_i = masss.begin(); mass_i != end; ++mass_i)
		if (!(*mass_i).second->isReady())
			return (*mass_i).second;

	return spMASS();
}

const size_t CMMASS::numOfMASSs() const {
	return masss.size();
}

void CMMASS::finish() throw (logic_error) {
	spMASS unfinished = this->getUnfinishedMASS();
	if (unfinished)
		throw logic_error("CMMASS::finish() - the MASS \""
			+unfinished->getName()+"\" is not finished.\n");

	ready = true;
}

const bool CMMASS::isReady() const {
	return ready;
}

void CMMASS::initiate(pthread_mutex_t *mutex_, pthread_cond_t *cond_)  throw (logic_error) {
	if (!this->isReady())
		throw logic_error(string("CMMASS::initiate() -\n"
			"The MMASS isn't ready yet.\n"));

	
	spAgent_set agents;
	pthread_attr_t attr;
	pthread_attr_init(&attr);
	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);

	spMASS_map::const_iterator mass_i, m_end = this->endMASSs();
	for (mass_i = this->beginMASSs(); mass_i != m_end; ++mass_i) {
		spMASS mass = (*mass_i).second;

		spAgent_map::const_iterator agent_i, a_end = mass->endAgents();
		for (agent_i = mass->beginAgents(); agent_i != a_end; ++agent_i) {
			pthread_create(&threads[(*agent_i).second->getName()], &attr,
				_agentBehaviourThread, (void*)&((*agent_i).second));
			agents.insert((*agent_i).second);
		}
	}

	ThreadArgument *argument = new ThreadArgument(mutex_, cond_, agents, this->self);
	pthread_create(&this->envThread, &attr, _environmentBehaviourThread, (void*)argument);
	// _environmentBehaviourThread((void*) argument);

	pthread_attr_destroy(&attr);
	agents.clear();

	/* ESPERA PELAS THREADS TERMINAR, VOU REITRAR MAIS TARDE!!!!! <-----------*/
	/*
	std::map<const std::string, pthread_t>::const_iterator thread_i,
		t_end = threads.end();
	for (thread_i = threads.begin(); thread_i != t_end; ++thread_i)
		pthread_join((*thread_i).second, NULL);
	pthread_join(envThread, NULL);
	*/
}

volatile unsigned long CMMASS::getTurn() {
	return turn;
}

volatile bool CMMASS::finishThread() {
	return finish_thread;
}

void CMMASS::setFinishThread(bool b) {
	finish_thread = b;
}

std::map<const std::string, pthread_t>::iterator CMMASS::beginAgentThreads() {
	return threads.begin();
}

std::map<const std::string, pthread_t>::iterator CMMASS::endAgentThreads() {
	return threads.end();
}

volatile pthread_t& CMMASS::getEnvironmentThread() {
	return envThread;
}

void CMMASS::nextTurn() {
	turn++;
}

volatile pthread_mutex_t& CMMASS::getMutex() {
	return mutex;
}

volatile pthread_cond_t& CMMASS::getCond() {
	return cond;
}

bool CMMASS::act(spAgent agent, unsigned long turn_) {
	bool ret = false;
	spAction nullAction;

	pthread_mutex_lock(&const_cast<pthread_mutex_t>(agent->getMutex()));
	while ((!agent->finishThread()) && (agent->getPossibleAction() != nullAction) && (agent->getAction() == nullAction) && (!agent->actionPerformed()) && (this->turn <= turn_)) {
		pthread_cond_wait(&const_cast<pthread_cond_t>(agent->getCond()), &const_cast<pthread_mutex_t>(agent->getMutex()));
	}

	if ((agent->getAction() != nullAction) || (this->turn > turn_) || (agent->actionAccepted())) { // este d pra tirar!
		agent->getPossibleAction().reset();
		ret = true;
	}

	agent->setActionPerformed(false);
	pthread_mutex_unlock(&const_cast<pthread_mutex_t>(agent->getMutex()));
	

	return ret;
}

CMMASS::CMMASS() {
	int rc;
	LIMIT = 1000;
	turn = 0;
	ready = false;
	finish_thread = false;
	rc = pthread_mutex_init(&const_cast<pthread_mutex_t>(mutex), NULL);
	rc += pthread_cond_init(&const_cast<pthread_cond_t>(cond), NULL);
	if (rc != 0) printf("CMMASS::CMMASS() ln110\n");
}

void MMASS::_clearOccupiedSites(spMMASS mmass) {
	spMASS_map::const_iterator mass_i, m_end = mmass->endMASSs();
	for (mass_i = mmass->beginMASSs(); mass_i != m_end; ++mass_i)
		(*mass_i).second->clearOccupiedSites();
}

void MMASS::_clearFields(spMMASS mmass) {
	spMASS_map::const_iterator mass_i, m_end = mmass->endMASSs();
	for (mass_i = mmass->beginMASSs(); mass_i != m_end; ++mass_i) {
		spSpace space = (*mass_i).second->getSpace();
		spSite_map::const_iterator site_i, s_end = space->endSites();
		for (site_i = space->beginSites(); site_i != s_end; ++site_i) {
			(*site_i).second->clearFields();
		}
	}
}

void MMASS::_performActions(spAgent_set &agents) {
	spAgent_set::iterator agent_i, a_end = agents.end();
	for (agent_i = agents.begin(); agent_i != a_end; ++agent_i) {
		pthread_mutex_lock(&const_cast<pthread_mutex_t>((*agent_i)->getMutex()));
		spAction action = (*agent_i)->getAction();
		action->performAction((*agent_i));
		(*agent_i)->setActionToNull();
		(*agent_i)->nextTurn();
		//printf("<font color=\"#006600\">MMASS  - %lu - _performActions - \"%s\" action is null</font><br>\n", (*agent_i)->getSite()->getSpace()->getMASS()->getMMASS()->getTurn(), (*agent_i)->getName().c_str()); fflush(NULL); // este d pra tirar!
		pthread_mutex_unlock(&const_cast<pthread_mutex_t>((*agent_i)->getMutex()));
	}
}

bool MMASS::_triggerManagement(spAgent agent, spAction action) {
	pthread_mutex_lock(&const_cast<pthread_mutex_t>(agent->getMutex()));
	agent->performAction();
	agent->setActionAccepted(true);
	agent->setActionPerformed(true);
	pthread_mutex_unlock(&const_cast<pthread_mutex_t>(agent->getMutex()));
	return true;
}

bool MMASS::_transportManagement(spAgent agent, spAction action, unsigned long turn) {
	spMASS mass = agent->getMASS();
	set<const string> occupiedSites = mass->getOccupiedSites();
	const string currentSite = agent->getSite()->getName();
	const string destinationSite = action->getDestinationSiteName();

	set<const string>::iterator site = occupiedSites.find(destinationSite);
	if (site != occupiedSites.end()) {
		pthread_mutex_lock(&const_cast<pthread_mutex_t>(agent->getMutex()));
		agent->addDeniedAction(action->createActionInfo());
		//printf("<font color=\"#006600\">MMASS  - %lu - _transportManagement - <ins>action denied</ins> for \"%s\", action identifier is %s</font><br>\n", turn, agent->getName().c_str(), action->createActionInfo()->getIdentifier().c_str()); fflush(NULL);
		agent->notifyFailure(turn);
		agent->setActionAccepted(false);
		agent->setActionPerformed(true);
		pthread_mutex_unlock(&const_cast<pthread_mutex_t>(agent->getMutex()));
		return false;
	}

	pthread_mutex_lock(&const_cast<pthread_mutex_t>(agent->getMutex()));
	agent->performAction();
	mass->eraseOccupiedSite(currentSite); // Maybe it can be removed (1)
	mass->addOccupiedSite(destinationSite);
	agent->setActionAccepted(true);
	agent->setActionPerformed(true);
	pthread_mutex_unlock(&const_cast<pthread_mutex_t>(agent->getMutex()));
	return true;
}

bool MMASS::_emitManagement(spAgent agent, spAction action, unsigned long turn) {
	pthread_mutex_lock(&const_cast<pthread_mutex_t>(agent->getMutex()));
	agent->performAction();
	agent->setActionAccepted(true);
	agent->setActionPerformed(true);
	pthread_mutex_unlock(&const_cast<pthread_mutex_t>(agent->getMutex()));
	return true;
}

bool MMASS::_reactionManagement(spAgent agent, spAction action, unsigned long turn) {
	spMASS mass = agent->getSite()->getSpace()->getMASS();
	bool agreed = true;
	spAgentInfo_set involvedAgents = action->getReactionPartners();
	spAgent_set reactingAgents;
	reactingAgents.insert(agent);

	spAgentInfo_set::iterator info_i, i_end = involvedAgents.end();
	for (info_i = involvedAgents.begin(); info_i != i_end; ++info_i) {
		spAgent agent = mass->getAgent((*info_i)->getAgentName());
		pthread_mutex_lock(&const_cast<pthread_mutex_t>(agent->getMutex()));
		if (!agent->agreeReaction(involvedAgents)) {
			pthread_mutex_unlock(&const_cast<pthread_mutex_t>(agent->getMutex()));
			agreed = false;
			break;
		}
		pthread_mutex_unlock(&const_cast<pthread_mutex_t>(agent->getMutex()));

		reactingAgents.insert(agent);
	}

	spAgent_set::iterator agent_i, a_end = reactingAgents.end();
	if (agreed) {
		spAction nullAction;
		for (agent_i = reactingAgents.begin(); agent_i != a_end; ++agent_i) {
			pthread_mutex_lock(&const_cast<pthread_mutex_t>((*agent_i)->getMutex()));
			if ((*agent_i)->getAction() == nullAction)
				(*agent_i)->setActionPerformed(true);
			else if ((*agent_i)->getAction()->getIdentifier() == "Transport") {
				const string currentSite = (*agent_i)->getSite()->getName();
				const string destinationSite = (*agent_i)->getAction()->getDestinationSiteName();
				mass->eraseOccupiedSite(destinationSite);
				mass->addOccupiedSite(currentSite); // Maybe it can be removed (1)
			}
			(*agent_i)->performAction();
			(*agent_i)->setActionAccepted(true);
			pthread_mutex_unlock(&const_cast<pthread_mutex_t>((*agent_i)->getMutex()));
		}
	} else {
		for (agent_i = reactingAgents.begin(); agent_i != a_end; ++agent_i) {
			pthread_mutex_lock(&const_cast<pthread_mutex_t>((*agent_i)->getMutex()));
			(*agent_i)->setActionAccepted(false);
			(*agent_i)->addDeniedAction(action->createActionInfo());
			(*agent_i)->notifyFailure(turn);
			pthread_cond_signal(&const_cast<pthread_cond_t>((*agent_i)->getCond()));
			pthread_mutex_unlock(&const_cast<pthread_mutex_t>((*agent_i)->getMutex()));
		}
		pthread_mutex_lock(&const_cast<pthread_mutex_t>(agent->getMutex()));
		agent->setActionPerformed(true);
		pthread_mutex_unlock(&const_cast<pthread_mutex_t>(agent->getMutex()));
		return false;
	}

	return true;
}

bool MMASS::_manage(spAgent agent, spAction action) {
	spAction nullAction;
	bool ret = true;

	pthread_mutex_lock(&const_cast<pthread_mutex_t>(agent->getMutex()));
	if ((agent->actionPerformed() && agent->actionAccepted() && (agent->getTurn() > turn)) || (agent->getAction() != nullAction)) { // if ((agent->actionPerformed() && agent->actionAccepted()) || (agent->getTurn() > turn))
		//printf("<font color=\"#006600\">MMASS  - %lu - _manage - <ins>no manage needed</ins> for \"%s\" (<font color=\"#0000ff\">signal</font>)</font><br>\n", turn, agent->getName().c_str()); fflush(NULL);
		pthread_cond_signal(&const_cast<pthread_cond_t>(agent->getCond()));
		pthread_mutex_unlock(&const_cast<pthread_mutex_t>(agent->getMutex()));
		return true;
	}
	pthread_mutex_unlock(&const_cast<pthread_mutex_t>(agent->getMutex()));

	if (action == nullAction) {
		//printf("<font color=\"#006600\">MMASS  - %lu - _manage - <ins>no action selected</ins> for \"%s\"</font><br>\n", turn, agent->getName().c_str()); fflush(NULL);
		return false;
	}

	ret = action->manage(agent, action, turn);

	pthread_mutex_lock(&const_cast<pthread_mutex_t>(agent->getMutex()));
	//printf("<font color=\"#006600\">MMASS  - %lu - _manage - managing done for \"%s\" (<font color=\"#0000ff\">signal</font>)</font><br>\n", turn, agent->getName().c_str()); fflush(NULL);
	pthread_cond_signal(&const_cast<pthread_cond_t>(agent->getCond()));
	pthread_mutex_unlock(&const_cast<pthread_mutex_t>(agent->getMutex()));

	return ret;
}

void* MMASS::_environmentBehaviourThread(void *arg) {
	ThreadArgument *argument = (ThreadArgument*)arg;
	spAgent_set agents = argument->agents;
	spMMASS mmass = argument->mmass;

	FILE *fp = fopen("output.txt", "a+b");
    if( !fp )
		return NULL;

	//fprintf(fp, "environment thread started\n"); fflush(fp);
	
	//printf("<font color=\"#006600\">MMASS  - %lu</font><br>\n", mmass->getTurn()); fflush(NULL);
	while (true) {
		spAgent_set agentsChoosingAction = agents;

		while (!agentsChoosingAction.empty()) {
			spAgent_set::iterator agent_i = agentsChoosingAction.begin(),
				a_end = agentsChoosingAction.end();

			while (agent_i != a_end) {
				unsigned long environmentTurn = mmass->getTurn();

				pthread_mutex_lock(&const_cast<pthread_mutex_t>((*agent_i)->getMutex()));
				spAction action = (*agent_i)->collectAction(environmentTurn);

				if ((*agent_i)->getTurn() >= environmentTurn) { // este no d pra tirar
					pthread_mutex_unlock(&const_cast<pthread_mutex_t>((*agent_i)->getMutex()));
					bool actionAccepted = _manage((*agent_i), action, environmentTurn);

					pthread_mutex_lock(&const_cast<pthread_mutex_t>((*agent_i)->getMutex()));
					if (actionAccepted) {
						pthread_mutex_unlock(&const_cast<pthread_mutex_t>((*agent_i)->getMutex()));
						agent_i = agentsChoosingAction.erase(agent_i);
						continue;
					}
				}
				//printf("<font color=\"#006600\">MMASS  - %lu - \"%s\" (turn is %d)</font><br>\n", environmentTurn, (*agent_i)->getName().c_str(), (*agent_i)->getTurn()); fflush(NULL); // este no d pra tirar
				pthread_mutex_unlock(&const_cast<pthread_mutex_t>((*agent_i)->getMutex()));
				++agent_i;
			}
		}

		pthread_mutex_lock(&const_cast<pthread_mutex_t>(mmass->getMutex()));
		_clearFields(mmass);
		_performActions(agents);
		_clearOccupiedSites(mmass);
		//printSitesState(mmass);
		mmass->nextTurn();
		//printInfo(agents);
		//printf("<font color=\"#006600\">MMASS  - %lu (<font color=\"#00ff00\">signal</font>)</font><br>\n", mmass->getTurn()); fflush(NULL);
		pthread_cond_broadcast(&const_cast<pthread_cond_t>(mmass->getCond()));
		pthread_mutex_unlock(&const_cast<pthread_mutex_t>(mmass->getMutex()));

		if (mmass->finishThread()) {
			fprintf(fp, "environment thread finished!\n"); fflush(fp);
			break;
		}
	}

	pthread_mutex_lock(&const_cast<pthread_mutex_t>(mmass->getMutex()));
	spAgent_set::const_iterator agent_i, a_end = agents.end();
	for (agent_i = agents.begin(); agent_i != a_end; ++agent_i) {
		pthread_mutex_lock(&const_cast<pthread_mutex_t>((*agent_i)->getMutex()));
		(*agent_i)->setFinishThread(true);
		pthread_cond_signal(&const_cast<pthread_cond_t>((*agent_i)->getCond()));
		pthread_mutex_unlock(&const_cast<pthread_mutex_t>((*agent_i)->getMutex()));
	}
	pthread_cond_broadcast(&const_cast<pthread_cond_t>(mmass->getCond()));
	pthread_mutex_unlock(&const_cast<pthread_mutex_t>(mmass->getMutex()));

	delete (ThreadArgument*)arg;
	pthread_exit(NULL);
	return NULL;
}

void MMASS::printInfo(spAgent_set &agents) {
	spAgent_set::const_iterator agent_i, a_end = agents.end();
	for (agent_i = agents.begin(); agent_i != a_end; ++agent_i) {
		if ((*agent_i)->getMASS()->getName() == "mass1")
			if ((*agent_i)->getName() == "agent1.3")
				printf("%s - %s<br>\n", (*agent_i)->getName().c_str(), (*agent_i)->getSite()->getName().c_str()); fflush(NULL);
	}
	printf("<br>\n");
}

void MMASS::printSitesState(spMMASS mmass) {
	set<const string> s;
	spMASS_map::const_iterator mass_i, m_end = mmass->endMASSs();
	for (mass_i = mmass->beginMASSs(); mass_i != m_end; ++mass_i) {
		spSpace space = (*mass_i).second->getSpace();
		spSite_map::const_iterator site_i, s_end = space->endSites();
		for (site_i = space->beginSites(); site_i != s_end; ++site_i) {
			spSite site = (*site_i).second;
			if ((*mass_i).second->getName() == "mass1") {
				printf("%s - FieldAgent = ", site->getName().c_str()); fflush(NULL);
				if (site->fields.count("FieldAgent") > 0) { printf("%7.5g - FieldAgent007 - ", site->fields["FieldAgent"]->getValue()); fflush(NULL); }
				else { printf ("%7.5g - FieldAgent007 - ", (double)0); fflush(NULL); }
				if (site->fields.count("FieldAgent007") > 0) { printf ("%7.5g", site->fields["FieldAgent007"]->getValue()); fflush(NULL); }
				else { printf("%7.5g", (double)0); fflush(NULL); }
				spAgent agent = site->getAgent();
				if (agent) s.insert(agent->getName());
				if (agent) { printf(" - %s", agent->getName().c_str()); fflush(NULL); }
				printf("\n"); fflush(NULL);
			}
		}
	}
	printf("\n"); fflush(NULL);
	if (s.size() != 3) {
		if (s.count("agent1.2") == 0) {
			spSite site = mmass->getMASS("mass1")->getAgent("agent1.2")->getSite();
			printf("agent1.2 is in %s\n", site->getName().c_str()); fflush(NULL);
			spAgent agent = site->getAgent();
			printf("%s has ", site->getName().c_str()); fflush(NULL);
			if (agent) { printf("%s\n", agent->getName().c_str()); fflush(NULL); }
			else { printf("no agent\n"); fflush(NULL); }
		}
		printf("failed! aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa  aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa  aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa  aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa  aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa  aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa  aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa \n");
	}
}

spMMASS CMMASS::self = spMMASS(new CMMASS);