#ifndef SPACE_H
#define SPACE_H
#include "MMASSDefs.h"
#include <string>
#include <set>
#include <stdexcept>

namespace MMASS {
	using std::string;

	class Space {
	public:
		class logic_error : public std::logic_error {
		public:
			logic_error(const string& msg = "") : std::logic_error(msg) {}
		};

		class invalid_argument : public std::invalid_argument {
		public:
			invalid_argument(const string& msg = "") : std::invalid_argument(msg) {}
		};

	protected:
		spSite_map sites; // sites
		// edges (adjacent relation)

	public:
		~Space();

		spSite createSite(const string &name_) throw (logic_error, invalid_argument);
		spSite createSite(const string &name_, spSite_set &adjSites) throw (logic_error, invalid_argument);

		const string& getName() const;

		const spSpace getSelf() const;

		const bool hasMASS() const;
		const spMASS getMASS() const;

		const bool hasSite() const;
		const bool hasSite(const string &name_) const;
		const bool hasSite(spSite site) const;
		const size_t numOfSites() const;
		spSite getSite(const string &name_) throw (logic_error, invalid_argument);
		spSite_map::const_iterator beginSites() const;
		spSite_map::const_iterator endSites() const;

		const bool isReady() const;
		void finish() throw (logic_error);
		
	private:
		const string name;
		wpSpace      self;
		bool         ready;
		wpMASS       mass;

		Space();

		static spSpace createSpace();

		void setSelfPointer(wpSpace self_);

		void setMASS(wpMASS mass_);

		void checkGraph() throw (logic_error);
		void dfsRcheck(spSite site);

		/* Hiding methods. */
		Space(const Space&);
		Space& operator=(const Space&);

		friend MASS;
	};
}

#endif