#include "stdafx.h"
#include "MASS.h"
#include "Site.h"
#include "Space.h"
using namespace MMASS;

MASS::~MASS() { }

spSpace MASS::createSpace() throw (logic_error) {
	// TODO: CRIAR UM SPACE AUTOMATICAMENTE!!!!!!!!!!!
	if (space)
		throw logic_error("MASS::createSpace() - "
			"The MASS \""+name+"\" already has a space.\n");

	spSpace space_ = Space::createSpace();
	space_->setMASS(self);
	space = space_;
	return space;
}

spAgent MASS::createAgent(Agent::AgentFactory &factory, const string &name_)
                          throw (invalid_argument, logic_error) {
	if (ready)
		throw logic_error(string("MASS::createAgent(Agent::AgentFactory &factory, "
			"const string &name_) - factory: ")+typeid(factory).name()+", name_: "
			+name_+"\n"
			"This MASS \""+name+"\" is ready. Can't create an agent with a finished "
			"MASS.\n");

	if (!space->isReady())
		throw logic_error(string("MASS::createAgent(Agent::AgentFactory &factory, "
			"const string &name_) - factory: ")+typeid(factory).name()+", name_: "
			+name_+"\n"
			"The space of this MASS \""+name+"\" isn't ready. Can't "
			"create an agent with a MASS with a unfinished space.\n");

	if (space->numOfSites() <= this->numOfAgents())
		throw logic_error(string("MASS::createAgent(Agent::AgentFactory &factory, "
			"const string &name_) - factory: ")+typeid(factory).name()+", name_: "
			+name_+"\n"
			" You are trying to create more agents than sites in this MASS \""
			+name+"\". Can't create more agents than sites in a MASS.\n");

	if (this->hasAgent(name_))
		throw invalid_argument(string("MASS::createAgent(Agent::AgentFactory &factory, "
			"const string &name_) - factory: ")+typeid(factory).name()+", name_: "
			+name_+"\n"
			"A agent with this name already exists in the MASS \""+name+
			"\".\n");

	spAgent agent = Agent::createAgent(factory, name_);
	agent->setMASS(self);
	agents[name_] = agent;

	return agent;
}

spAgent MASS::createAgent(Agent::AgentFactory &factory, const string &agent_name,
						  const string site_name) throw (invalid_argument, logic_error) {
	if (ready)
		throw logic_error(string("MASS::createAgent(Agent::AgentFactory &factory, "
			"const string &agent_name, const string site_name) - factory: ")
			+typeid(factory).name()+", agent_name: "+agent_name+", site_name: "
			+site_name+"\n"
			"This MASS \""+name+"\" is ready. Can't create an agent with a finished "
			"MASS.\n");

	if (!space->isReady())
		throw logic_error(string("MASS::createAgent(Agent::AgentFactory &factory, "
			"const string &agent_name, const string site_name) - factory: ")
			+typeid(factory).name()+", agent_name: "+agent_name+", site_name: "
			+site_name+"\n"
			"The space of this MASS \""+name+"\" isn't ready. Can't "
			"create an agent with a MASS with a unfinished space.\n");

	if (space->numOfSites() <= this->numOfAgents())
		throw logic_error(string("MASS::createAgent(Agent::AgentFactory &factory, "
			"const string &agent_name, const string site_name) - factory: ")
			+typeid(factory).name()+", agent_name: "+agent_name+", site_name: "
			+site_name+"\n"
			" You are trying to create more agents than sites in this MASS \""
			+name+"\". Can't create more agents than sites in a MASS.\n");

	if (this->hasAgent(agent_name))
		throw invalid_argument(string("MASS::createAgent(Agent::AgentFactory &factory, "
			"const string &agent_name, const string site_name) - factory: ")
			+typeid(factory).name()+", agent_name: "+agent_name+", site_name: "
			+site_name+"\n"
			"An agent with this name already exists in the MASS \""+name+
			"\".\n");

	spSite site = space->getSite(site_name);
	if (site->hasAgent())
		throw invalid_argument(string("MASS::createAgent(Agent::AgentFactory &factory, "
			"const string &agent_name, const string site_name) - factory: ")
			+typeid(factory).name()+", agent_name: "+agent_name+", site_name: "
			+site_name+"\n"
			"The site \""+site->getName()+"\" already has the agent \""
			+site->getAgent()->getName()+"\" on it.\n");

	spAgent agent = Agent::createAgent(factory, agent_name);
	agent->setMASS(self);

	try {
		agent->setSite(site);
	} catch (Agent::invalid_argument &e) {
		throw invalid_argument(string("MASS::createAgent(Agent::AgentFactory &factory, "
			"const string &agent_name, const string site_name) - factory: ")
			+typeid(factory).name()+", agent_name: "+agent_name+", site_name: "
			+site_name+"\n"
			"<-- "+e.what());
	}

	agents[agent_name] = agent;

	return agent;
}

const string& MASS::getName() const {
	return name;
}

const spMASS MASS::getSelf() const {
	return self.lock();
}

const spMMASS MASS::getMMASS() const {
	return mmass.lock();
}

const bool MASS::hasAgent() const {
	return (!agents.empty());
}

const bool MASS::hasAgent(const std::string &name_) const {
	return (agents.count(name_) > 0);
}

const bool MASS::hasAgent(const spAgent agent) const {
	return (this->hasAgent(agent->getName()));
}

spAgent_map::const_iterator MASS::beginAgents() const {
	return agents.begin();
}

spAgent_map::const_iterator MASS::endAgents() const {
	return agents.end();
}

const spAgent MASS::getAgentWithoutSite() const throw (logic_error) {
	spAgent_map::const_iterator agent_i, end = agents.end();
	for (agent_i = agents.begin(); agent_i != end; ++agent_i) {
		if (!(*agent_i).second->hasSite()) return (*agent_i).second;
	}

	return spAgent();
}

const size_t MASS::numOfAgents() const {
	return agents.size();
}

spAgent MASS::getAgent(const string &name_) throw (logic_error, invalid_argument) {
	if (!space)
		throw logic_error("MASS::getAgent(const string &name_) - "
			"name_: "+name_+"\n"
			"This MASS \""+name+"\" doesn't have a "
			"space.\n");

	if (!space->isReady())
		throw logic_error("MASS::getAgent(const string &name_) - "
			"name_: "+name_+"\n"
			"This MASS \""+name+"\" has a space that isn't"
			" ready yet.\n");

	if (!this->hasAgent())
		throw logic_error("MASS::getAgent(const string &name_) - "
			"name_: "+name_+"\n"
			"This MASS \""+name+"\" doesn't have any"
			" agent yet.\n");

	if (!this->hasAgent(name_))
		throw invalid_argument("MASS::getAgent(const string &name_) - "
			"name_: "+name_+"\n"
			"This MASS \""+name+"\" doesn't have an"
			" agent with this name.\n");

	return agents[name_];
}

const bool MASS::hasSpace() const {
	return space;
}

const spSpace MASS::getSpace() const {
	return space;
}

void MASS::finish() throw (logic_error) {
	if (!space)
		throw logic_error("MASS::finish() - this MASS \""+name+"\""
			" doesn't have a space.\n");

	if (!space->isReady())
		throw logic_error("MASS::finish() - this MASS \""+name+"\""
			" doesn't have a finished space.\n");

	if (!this->hasAgent())
		throw logic_error("MASS::finish() - this MASS \""+name+"\""
			" doesn't have an agent.\n");

	spAgent lostAgent = this->getAgentWithoutSite();
	if (lostAgent)
		throw logic_error("MASS::finish() - the agent \""
			+lostAgent->getName()+"\", from MASS \""+name+"\""
			", isn't on a site.\n");

	ready = true;
}

const bool MASS::isReady() const {
	return ready;
}

spMASS MASS::createMASS(const string &name_) {
	spMASS mass(new MASS(name_));
	mass->setSelfPointer(mass);

	return mass;
}

MASS::MASS(const string &name_) : name(name_) { ready = false; }

void MASS::setSelfPointer(wpMASS self_) {
	self = self_;
}

void MASS::setMMASS(wpMMASS mmass_) {
	mmass = mmass_;
}

set<const string>& MASS::getOccupiedSites() {
	if (occupiedSites.empty()) {
		spAgent_map::iterator agent_i, a_end = agents.end();
		for (agent_i = agents.begin(); agent_i != a_end; ++agent_i)
			occupiedSites.insert((*agent_i).second->getSite()->getName());
	}

	return occupiedSites;
}

void MASS::addOccupiedSite(const string &name_) {
	occupiedSites.insert(name_);
}

void MASS::eraseOccupiedSite(const string &name_) {
	occupiedSites.erase(name_);
}

void MASS::clearOccupiedSites() {
	occupiedSites.clear();
}