#include "TestAgent.h"
#include "Agent007.h"
#include "../MMASSdll/Site.h"
#include "../MMASSdll/Space.h"
#include "../MMASSdll/MASS.h"
#include "../MMASSdll/MMASS.h"


void TestAgent::run() {
	usingAgentTests();
}

void TestAgent::usingAgentTests() {
	try { /* Putting an agent on a empty site. This agent wasn't
		  previously on a site. */
		spMASS mass1 = CMMASS::getSingleton()->getMASS("mass1");
		spAgent agent = mass1->getAgent("agent1.1");
		spSite site = mass1->getSpace()->getSite("site1.2");

		test_(!agent->hasSite());
		test_(!site->hasAgent());

		agent->setSite(site);

		test_(agent->getSite() == site);
		test_(site->getAgent() == agent);
	} catch (std::exception &e) {
		fail_(string("An unexpected exception have occured:\n")
			+ e.what());
	}

	try { /* Putting an agent on a empty site. This agent was
		  on a site already. */
		spMASS mass1 = CMMASS::getSingleton()->getMASS("mass1");
		spAgent agent = mass1->getAgent("agent1.1");
		spSite site = mass1->getSpace()->getSite("site1.3");

		test_(agent->hasSite());
		test_(!site->hasAgent());

		spSite oldSite = agent->getSite();
		test_(oldSite->getAgent() == agent);

		agent->setSite(site);

		test_(agent->getSite() == site);
		test_(site->getAgent() == agent);
		test_(!oldSite->hasAgent());
	} catch (std::exception &e) {
		fail_(string("An unexpected exception have occured:\n")
			+ e.what());
	}

	try { /* Can't put an agent on a occupied site. This agent wasn't
		  previously on a site. */
		spMASS mass1 = CMMASS::getSingleton()->getMASS("mass1");
		spAgent agent = mass1->getAgent("agent1.3");
		spSite site = mass1->getSpace()->getSite("site1.3");

		test_(!agent->hasSite());
		test_(site->hasAgent());
		test_(site->getAgent() != agent);

		agent->setSite(site);

		fail_("This site already has an agent on it."
			"It should have thrown an exception.");
	} catch (Agent::invalid_argument&) {
		succeed_();

		spMASS mass1 = CMMASS::getSingleton()->getMASS("mass1");
		spAgent agent = mass1->getAgent("agent1.3");

		test_(!agent->hasSite());
	} catch (std::exception &e) {
		fail_(string("An unexpected exception have occured:\n")
			+ e.what());
	}

	try { /* Can't put an agent on a occupied site. This agent was
		  previously on a site. */
		spMASS mass1 = CMMASS::getSingleton()->getMASS("mass1");
		spAgent agent = mass1->getAgent("agent1.2");
		spSite site = mass1->getSpace()->getSite("site1.3");

		test_(agent->hasSite());
		test_(site->hasAgent());
		test_(site->getAgent() != agent);

		agent->setSite(site);

		fail_("This site already has an agent on it."
			"It should have thrown an exception.");
	} catch (Agent::invalid_argument&) {
		succeed_();

		spMASS mass1 = CMMASS::getSingleton()->getMASS("mass1");
		spAgent agent = mass1->getAgent("agent1.2");

		test_(agent->hasSite());
	} catch (std::exception &e) {
		fail_(string("An unexpected exception have occured:\n")
			+ e.what());
	}

	try { /* Putting an agent on its own site doesn't have any
		  effect. */
		spMASS mass1 = CMMASS::getSingleton()->getMASS("mass1");
		spAgent agent = mass1->getAgent("agent1.1");
		spSite site = mass1->getSpace()->getSite("site1.3");

		test_(agent->hasSite());
		test_(site->hasAgent());
		test_(site->getAgent() == agent);

		agent->setSite(site);

		succeed_();
	} catch (std::exception &e) {
		fail_(string("An unexpected exception have occured:\n")
			+ e.what());
	}

	try { /* Can't set an agent's site to null. This agent wasn't
		  previously on a site. */
		spMASS mass1 = CMMASS::getSingleton()->getMASS("mass1");
		spAgent agent = mass1->getAgent("agent1.3");
		spSite nullSite;

		test_(!agent->hasSite());

		agent->setSite(nullSite);

		fail_("This site is null."
			"It should have thrown an exception.");
	} catch (Agent::invalid_argument&) {
		succeed_();

		spMASS mass1 = CMMASS::getSingleton()->getMASS("mass1");
		spAgent agent = mass1->getAgent("agent1.3");

		test_(!agent->hasSite());
	} catch (std::exception &e) {
		fail_(string("An unexpected exception have occured:\n")
			+ e.what());
	}

	try { /* Can't set an agent's site to null. This agent was
		  previously on a site. */
		spMASS mass1 = CMMASS::getSingleton()->getMASS("mass1");
		spAgent agent = mass1->getAgent("agent1.2");
		spSite nullSite;

		test_(agent->hasSite());

		agent->setSite(nullSite);

		fail_("This site is null."
			"It should have thrown an exception.");
	} catch (Agent::invalid_argument&) {
		succeed_();

		spMASS mass1 = CMMASS::getSingleton()->getMASS("mass1");
		spAgent agent = mass1->getAgent("agent1.2");

		test_(agent->hasSite());
	} catch (std::exception &e) {
		fail_(string("An unexpected exception have occured:\n")
			+ e.what());
	}

	try { /* Can't set an agent's site to a site that isn't
		  from the same agent's MASS. */
		spMASS mass1 = CMMASS::getSingleton()->getMASS("mass1");
		spSite site = mass1->getSpace()->getSite("site1.7");
		test_(!site->hasAgent());
		spMASS mass3 = CMMASS::getSingleton()->getMASS("mass3");
		spAgent agent = mass3->getAgent("agent3.1");

		test_(site->getSpace()->getMASS() != agent->getMASS());

		agent->setSite(site);

		fail_("This site isn't from this agent's MASS."
			"It should have thrown an exception.");
	} catch (Agent::logic_error&) {
		succeed_();

		spMASS mass1 = CMMASS::getSingleton()->getMASS("mass1");
		spSite site = mass1->getSpace()->getSite("site1.7");

		test_(!site->hasAgent());
	} catch (std::exception &e) {
		fail_(string("An unexpected exception have occured:\n")
			+ e.what());
	}
}
